#   SAP_DB.pm
#
#   Copyright (c) 2000-2001, SAP AG
#
#   You may distribute under the terms of the
#   GNU Library General Public License
#
#
#    ========== licence begin LGPL
#    Copyright (C) 2000-2001 SAP AG
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#    ========== licence end
#


use SAP::DBTech::sapdb;
use strict;

{
    package DBD::SAP_DB;
    require DBI;

    my $VERSION = "7.3";

    my $drh = undef;       # holds driver handle once initialised
    my $err = 0;           # The $DBI::err value

    sub driver{
        return $drh if $drh;
        my($class, $attr) = @_;
        $class .= "::dr";
        ($drh) = DBI::_new_drh($class, {
            'Name' => 'SAP_DB',
            'Version' => $VERSION,
            'Attribution' => 'SAP DB driver by SAP AG',
            }, [ qw'example implementors private data']);
        $drh;
    }

    sub parseError {
        my ($handle, $fullString) = @_;
        my ($errcode, $errtext) = (-1, "unknown error");
        if ($fullString =~ /SQLError: \((-?[0-9]+)\) (.*) at /) {
            $errcode = $1;
            $errtext = $2;
        }
        else {
            $errtext = $fullString;
        }
        $handle->DBI::set_err ($errcode, $errtext);
    }

}


{   package DBD::SAP_DB::dr; # ====== DRIVER ======
    $DBD::SAP_DB::dr::imp_data_size = 0;
    # we use default (dummy) connect method

    sub connect {
        my($drh, $dbname, $user, $auth, $attr)= @_;

        # Some database specific verifications, default settings
        # and the like following here. This should only include
        # syntax checks or similar stuff where it's legal to
        # 'die' in case of errors.

        # create a 'blank' dbh (call superclass constructor)
        my $dbh = DBI::_new_dbh($drh, {
            'Name' => $dbname,
            'USER' => $user,
            'CURRENT_USER' => $user,
        });

        # Process attributes from the DSN; we assume ODBC syntax
        # here, that is, the DSN looks like var1=val1;...;varN=valN
        my @connectParts = split(/;/, $dbname);

        # Parse the DB part:
        my $host = ''; # Local connection by default.
        $dbname = shift @connectParts;
        if ($dbname =~ /^([^:]+):(.*)/) {
            $host = $1;
            $dbname = $2;
        }

        # Parse the options part:
        my @options;
        foreach my $var (@connectParts) {
            if ($var =~ /(.*?)=(,*)/) {
                $dbh->STORE($1, $2); # Not !!! $dbh->{$var} = $val;
                push (@options, $var);
            }
        }

        my $sapdb_session = undef;
        eval {
            my $optionString = @options?join('&', @options):'';

            $sapdb_session = sapdb::connect ($user, $auth,
                                             $dbname, $host, $optionString);
            $dbh->STORE ('sapdb_session', $sapdb_session);
        };
        if ($@) {
            DBD::SAP_DB::parseError ($dbh, $@);
            return undef;
        }
        return $dbh;
    }

    sub disconnect_all { }
    sub DESTROY { undef }
}


{   package DBD::SAP_DB::db; # ====== DATABASE ======
    $DBD::SAP_DB::db::imp_data_size = 0;

    sub disconnect {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->release ();
    }

    sub commit {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->commit ();
    }

    sub rollback {
        my($dbh)= @_;
        $dbh->FETCH ('sapdb_session')->rollback ();
    }

    sub prepare {
        my($dbh, $statement)= @_;

        my $sapdb_prepared;
        eval {
            $sapdb_prepared = $dbh->FETCH ('sapdb_session')
                ->prepare ($statement);
        };
        if ($@) {
            DBD::SAP_DB::parseError ($dbh, $@);
            return undef;
        }
        my($outer, $sth) = DBI::_new_sth($dbh, {
            'Statement'     => $statement,
            }, [ qw'example implementors private data']);

        $sth->STORE ('sapdb_prepared', $sapdb_prepared);
        #
        # calculate in-out-mapping
        #
        my $pinfos = $sapdb_prepared->getDescription ();
        my $inCount = 0;
        my $outCount = 0;
        my @inMap;
        my @outMap;
        push @inMap, undef;
        push @outMap, undef;
        foreach my $pinfo (@{$pinfos}) {
            my ($name, $typeName, $typeId, $length, $frac, $maybeNull, $ioMode)
                = @{$pinfo};
            if ($ioMode eq "IN") {
                push @inMap, $inCount;
                push @outMap, undef;
                ++$inCount;
            }
            elsif ($ioMode eq "OUT") {
                push @inMap, undef;
                push @outMap, $outCount;
                ++$outCount;
            }
            elsif ($ioMode eq "IN/OUT") {
                push @inMap, $inCount;
                push @outMap, $outCount;
                ++$inCount;
                ++$outCount;
            }
            else {
                print "terribly wrong\n";
            }
        }
        #
        # store attributes
        #
        my $completelyUnnecessaryLocalVariable = @{$pinfos};
        #CAAL: how do I pass the length of @{$pinfos} in one expression?
        $sth->STORE ('NUM_OF_PARAMS', $completelyUnnecessaryLocalVariable);
        $sth->STORE ('sapdb_inmap', \@inMap);
        $sth->STORE ('sapdb_outmap', \@outMap);
        $sth->STORE ('sapdb_inparameters', []);
        $sth->STORE ('sapdb_outparameters', []);
        $outer;
    }

    sub tables {
        my ($dbh) = @_;
        my $cursor = $dbh->FETCH ('sapdb_session')->sql (
            'select owner, tablename from tables where tabletype <> \'RESULT\'');
        my @result;
        my ($row, $owner, $tablename, $entry);
        while (defined ($row = $cursor->next ())) {
            $entry = join ('.', @{$row});
            push (@result, $entry);
        }
        return @result;
    }

    sub table_info {
        my ($dbh) = @_;
        my $prepared = $dbh->prepare ('select ' .
            'NULL TABLE_QUALIFIER, ' .
            'owner TABLE_OWNER, ' .
            'tablename TABLE_NAME, ' .
            'tabletype TABLE_TYPE, ' .
            'comment REMARKS ' .
            ' from tables');
        $prepared->execute ();
        return $prepared;
    }

    sub FETCH {
        my ($dbh, $attrib) = @_;
        # In reality this would interrogate the database engine to
        # either return dynamic values that cannot be precomputed
        # or fetch and cache attribute values too expensive to prefetch.
        if ($attrib eq 'AutoCommit') {
            return $dbh->FETCH ('sapdb_session')->getAutocommit ();
        }
        # else pass up to DBI to handle
        return $dbh->DBD::_::db::FETCH($attrib);
        }

    sub STORE {
        my ($dbh, $attrib, $value) = @_;
        # would normally validate and only store known attributes
        # else pass up to DBI to handle
        if ($attrib eq 'AutoCommit') {
            $dbh->FETCH ('sapdb_session')->setAutocommit ($value);
            return 1;
        }
        if ($attrib =~ /^sapdb_/) {
            $dbh->{$attrib} = $value; # Yes, we are allowed to do this,
            return 1;                 # but only for our private attributes
        }
        return $dbh->DBD::_::db::STORE($attrib, $value);
    }

    sub DESTROY { undef }
}


{   package DBD::SAP_DB::st; # ====== STATEMENT ======
    $DBD::SAP_DB::st::imp_data_size = 0;

    sub execute {
        my ($sth, @args) = @_;
        my ($result, $parameters);

        $result = '0E0';
        eval {
            if (@args == 0) {
                my $inparameters = $sth->FETCH ('sapdb_inparameters');
                my @parmvalues;
                foreach my $pair (@{$inparameters}) {
                    my ($parm, $isRef) = @{$pair};
                    if ($isRef) {
                        push @parmvalues, $$parm;
                    }
                    else {
                        push @parmvalues, $parm;
                    }
                }
                $parameters = \@parmvalues;
            }
            else {
                $parameters = \@args;
            }
            $result = $sth->FETCH ('sapdb_prepared')->execute ($parameters);
        };

        # return 1 if $@ =~ /^SQLError: \(100\)/; # FF: This is not an error!

        if ($@) {
            DBD::SAP_DB::parseError ($sth, $@);
            return undef;
        }
        my $refresult = ref ($result);
        if (ref ($result) eq 'SapDB_ResultSet') {
            #
            # SELECT
            #
            $sth->STORE ('sapdb_cursor', $result);
            # set cursor information
            my $colCount = $result->columnCount ();
            $sth->STORE ('NUM_OF_FIELDS', $colCount);
            my $pinfos = $result->getDescription ();
            my ($i, $pinfo, @colnames, @colTypes, @colPrecision, @colScale, @colNullable);
            my ($name, $typeName, $typeId, $length, $frac, $nullable, $ioMode);
            foreach $pinfo (@{$pinfos}) {
                ($name, $typeName, $typeId, $length, $frac, $nullable, $ioMode)
                    = @{$pinfo};
                push (@colnames, $name);
                push (@colTypes, $typeId);
                push (@colPrecision, $length);
                push (@colScale, $frac);
                push (@colNullable, $nullable);
            }
            $sth->STORE ('sapdb_colnames', \@colnames);
            $sth->STORE ('sapdb_coltypes', \@colTypes);
            $sth->STORE ('sapdb_colprecision', \@colPrecision);
            $sth->STORE ('sapdb_colscale', \@colScale);
            $sth->STORE ('sapdb_colnullable', \@colNullable);
        }
        elsif (ref ($result) eq 'ARRAY') {
            #
            # output parameters
            #
            my $outparameters = $sth->FETCH ('sapdb_outparameters');
            for (my $i = 0; $i < @{$result}; ++$i) {
                my $outref = $outparameters->[$i];
                if (defined ($outref)) {
                    $$outref = $result->[$i];
                }
            }
        }
        else {
            #
            # result count
            #
            if ($result == 0) {
                $result = '0E0';
            }
        }
        return $result;
    }

    sub bind_param {
        my ($sth, $index, $value) = @_;
        my ($parameters, $inMap, $parameterCount);

        $inMap = $sth->FETCH ('sapdb_inmap');
        my $inIndex = $inMap->[$index];
        if (defined ($inIndex)) {
            $parameters = $sth->FETCH ('sapdb_inparameters');
            $parameters->[$inIndex] = [$value, undef];
        }
    }

    sub bind_col {
        my ($sth, $index, $reference) = @_;
        my ($colBindings) = getColBindings ($sth);

        $colBindings->[$index - 1] = $reference;
    }

    sub bind_columns {
        my ($sth, @references) = @_;
        my ($colBindings) = getColBindings ($sth);
        my ($i);

        if (! defined ($references[0])) {
            shift @references;
        }
        for ($i = 0; $i < @references; ++$i) {
            $colBindings->[$i] = $references[$i];
        }
    }

    sub getColBindings {
        my ($sth) = @_;
        my $colBindings = $sth->FETCH ('sapdb_colbindings');

        if (!defined ($colBindings)) {
            $colBindings = [];
            my $cursor = $sth->FETCH ('sapdb_cursor');
            my $colCount = $cursor->columnCount ();
            for (my $i = 0; $i < $colCount; ++$i) {
                push (@{$colBindings}, undef);
            }
            $sth->STORE ('sapdb_colbindings', $colBindings);
        }
        return $colBindings;
    }

    sub bind_param_inout {
        my ($sth, $index, $value) = @_;
        my ($parameters, $inMap, $outMap, $parameterCount);

        $inMap = $sth->FETCH ('sapdb_inmap');
        my $inIndex = $inMap->[$index];
        if (defined ($inIndex)) {
            $parameters = $sth->FETCH ('sapdb_inparameters');
            $parameters->[$inIndex] = [$value, 1];
        }
        $outMap = $sth->FETCH ('sapdb_outmap');
        my $outIndex = $outMap->[$index];
        if (defined ($outIndex)) {
            $parameters = $sth->FETCH ('sapdb_outparameters');
            $parameters->[$inIndex] = ($value);
        }
    }

    sub fetch {
        my($sth) = @_;

        my $cursor = $sth->FETCH ('sapdb_cursor');
        my $row = undef;
        eval {
            $row = $cursor->next() if defined $cursor; # FF: it might not be (empty retsult)
        };
        if ($@) {
            DBD::SAP_DB::parseError ($sth, $@);
            return undef;
        }
        if (!$row) {
            $sth->finish ();     # no more data so finish
        }
        if ($row) {
            my $colBindings = $sth->FETCH ('sapdb_colbindings');
            if (defined ($colBindings)) {
                my ($i, $reference);
                for ($i = 0; $i < @{$row}; ++$i) {
                    $reference = $colBindings->[$i];
                    if (defined ($reference)) {
                        $$reference = $row->[$i];
                    }
                }
            }
        }
        return $row;
    }

    sub fetchrow_hashref {
        my($sth) = @_;
        my $row = $sth->fetch();
        return undef unless $row;

        my $names = $sth->{'sapdb_colnames'};
        my @namearray = @{$names};
        my %result;
        for (my $i = 0; $i < @namearray; ++$i) {
            $result{$namearray[$i]} = $row->[$i];
        }
        return \%result;
    }

    sub fetchrow_arrayref {
        my($sth) = @_;
        return $sth->DBI::st::fetch();
    }

    sub finish {
        my($sth) = @_;
    }

    sub FETCH {
        my ($sth, $attrib) = @_;
        # would normally validate and only fetch known attributes
        # else pass up to DBI to handle
        if ($attrib =~ /^sapdb_/) {
            return $sth->{$attrib};
        } elsif ($attrib eq 'NAME') {
            return $sth->{'sapdb_colnames'};
        } elsif ($attrib eq 'TYPE') {
            return $sth->{'sapdb_coltypes'};
        } elsif ($attrib eq 'PRECISION') {
            return $sth->{'sapdb_colprecision'};
        } elsif ($attrib eq 'SCALE') {
            return $sth->{'sapdb_colscale'};
        } elsif ($attrib eq 'NULLABLE') {
            return $sth->{'sapdb_colnullable'};
        } elsif ($attrib eq 'CursorName') {
            return $sth->{'sapdb_cursor'}->cursorName ();
        }
        return $sth->DBD::_::st::FETCH($attrib);
    }

    sub STORE {
        my ($sth, $attrib, $value) = @_;
        # would normally validate and only store known attributes
        # else pass up to DBI to handle

        # FF: This is just a quick hack to disallow setting NUM_OF_FIELDS more than once.
        if ($attrib eq 'NUM_OF_FIELDS') {
            if (exists $sth->{sapdb_NUM_OF_FIELDS}) {
            return 1;
            } else {
            $sth->{sapdb_NUM_OF_FIELDS} = 1;
            }
        }

        if ($attrib =~ /^sapdb_/) {
            $sth->{$attrib} = $value; # Yes, we are allowed to do this,
            return 1;                 # but only for our private attributes
        }
        return $sth->DBD::_::st::STORE($attrib, $value);
    }

    sub DESTROY { undef }
}

1;
__END__

=head1 NAME

DBD::SAP_DB - A DBI driver for SAP DB

=head1 SYNOPSIS

  use DBI;

  $dbh = DBI->connect("dbi:SAP_DN:databasename", $user, $passwd);
  $dbh = DBI->connect("dbi:SAP_DN:host:databasename", $user, $passwd);

  # See the DBI module documentation for full details

=head1 DESCRIPTION

To be written

=head1 SAP DB

Works with SAP DB 7.2 and higher.

  http://www.sapdb.org/

=head1 AUTHORS

DBD::SAP_DB is maintained as part of the SAP DB project.

Mail bug reports to sapdb-general@sap.com

=cut
